/*
 * =======================================================================
 * FILE:        PFIFO.c
 * DESCRIPTION: Pointer FIFO
 * CREATED      20 MAR 10
 * MODIFIED     
 * PROGRAMMER:  William L. Bahn
 * COPYRIGHT:   2010
 *
 * These functions may be used for any purpose without royalty provided
 * the original author is acknowledged.
 * =======================================================================
 */

/*
 * =======================================================================
 * REVISION HISTORY
 *
 * REV 1.1: 22 MAR 10 Modified to use get/set autogeneration
 * REV 1.0: 20 MAR 10 Initial Creation.
 * =======================================================================
 */

//===========================================================================

//---------------------------------------------------------------------------
// Include Files
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>

#include "PFIFO.h"
#include "dirtyd.h"
#include "bytes.h"

//---------------------------------------------------------------------------
// Structure Definition
//---------------------------------------------------------------------------

#define MAJOR_REV (1)
#define MINOR_REV (1)

#define STRUCT PFIFO
#define CODESTRING "PFIF"

struct STRUCT
{
	// Housekeeping Elements
	DWORD code;
	char *name;
	// Live Elements
	size_t depth; // 0
	size_t width; // 0
	size_t fill;  // 0
	size_t mask;  // 0 
	int empty;    // 0
	int full;     // 0
	void **queue; // NULL
	size_t read;  // 0
	size_t write; // 0
	void **block; // NULL
};

#define CODE ((((((\
	(CODESTRING[0])<<8)\
	+CODESTRING[1])<<8)\
	+CODESTRING[2])<<8)\
	+CODESTRING[3])

//===========================================================================
// Primitive Functions
//===========================================================================
//---------------------------------------------------------------------------
// Housekeeping
//---------------------------------------------------------------------------

DWORD f(get_code)(STRUCT *p)
{
	return (p)? p->code : 0;
}

DWORD f(set_code)(STRUCT *p)
{
	if (p)
		p->code = CODE;

	return f(get_code)(p);
}

int f(structcheck)(STRUCT *p)
{
	if ((p)&&(f(get_code)(p)==CODE))
		return 1;
	return 0;
}

char *f(get_name)(STRUCT *p)
{
	return (f(structcheck)(p))? p->name : NULL;
}

char *f(set_name)(STRUCT *p)
{
	if (f(structcheck)(p))
		p->name = VNQUOTES(STRUCT);
	return f(get_name)(p);
}

STRUCT *f(create)(void)
{
	STRUCT *p;

	p = (STRUCT *) malloc(sizeof(STRUCT));
	f(set_code)(p);
	f(set_name)(p);

	return p;
}

STRUCT *f(destroy)(STRUCT *p)
{
	if (f(structcheck)(p))
	{
		free(p);
		p = NULL;
	}

	return p;
}

//---------------------------------------------------------------------------
// Element Access
//---------------------------------------------------------------------------

getset(size_t, depth, 0);
getset(size_t, width, 0);
getset(size_t, fill,  0);
getset(size_t, mask,  0);
getset(int,    empty, 0);
getset(int,    full,  0);
getset(void**, queue, NULL);
getset(size_t, read,  0);
getset(size_t, write, 0);
getset(void**, block, NULL);

//===========================================================================
// Private Functions
//===========================================================================

// The private functions perform little to no parameter checking.
// The public functions calling them are responsible for this.

void f(report_elements)(STRUCT* p)
{
	printf("   depth = %u\n", f(get_depth)(p));
	printf("   mask  = %X\n", f(get_mask)(p));
	printf("   width = %u\n", f(get_width)(p));
	printf("   fill  = %u\n", f(get_fill)(p));
	printf("   empty = %u\n", f(get_empty)(p));
	printf("   full  = %u\n", f(get_full)(p));
	printf("   queue = %p\n", f(get_queue)(p));
	printf("   read  = %u\n", f(get_read)(p));
	printf("   write = %u\n", f(get_write)(p));
	printf("   block = %p\n", f(get_block)(p));
}

void copy(char *s, char *d, size_t bytes)
{
	size_t i;
	
	for (i = 0; i < bytes; i++)
		d[i] = s[i];
}

//===========================================================================
// Public Functions
//===========================================================================

WORD f(version)(int type)
{
	switch (type)
	{
		case STRUCT_VERSION: return (MAJOR_REV<<8)+MINOR_REV; break;
		case STRUCT_MAJOR:   return MAJOR_REV; break;
		case STRUCT_MINOR:   return MINOR_REV; break;
	}
	return 0;
}

void f(report)(STRUCT* p)
{
	printf("\n");
	printf("================================================\n");
	printf("Structure Name: %s (Code: %lX)\n", f(get_name)(p), f(get_code)(p));
	printf("Version: (%u.%u)\n", 
		f(version)(STRUCT_MAJOR), f(version)(STRUCT_MINOR));
	printf("------------------------------------------------\n");
	printf("   Size: %i bytes\n", sizeof(STRUCT));
	printf("   Addr: %p\n", p);
	printf("------------------------------------------------\n");
	f(report_elements)(p);
	printf("================================================\n");
}

//---------------------------------------------------------------------------

STRUCT* f(delete)(STRUCT *p)
{
	size_t i;

	if (f(structcheck)(p))
	{
		if (f(get_block)(p)) // Queue memory in one block
		{
			free1D(f(get_block)(p));
			f(set_block)(p, NULL);
		}
		else
		{
			for (i = 0; i < f(get_depth)(p); i++)
				if (f(get_queue)(p)[i])
				{
					free1D(f(get_queue)(p)[i]);
					f(get_queue)(p)[i] = NULL;
				}
		}
		p = f(destroy)(p);
	}

	return NULL;
}

STRUCT* f(new)(STRUCT *p, DWORD depth, DWORD width)
{
	size_t i, mask;

	if (p)
		if (f(structcheck)(p))
			p = f(delete)(p);
		else
			return NULL;

	// Sanity checks
	if (p) return NULL;
	if (0 == depth) return NULL;
	if (0 == width) return NULL;

	p = f(create)( );
	if (!f(structcheck)(p))
		return NULL;

    // The depth must be a power of 2 in order for bit-masking to work.
	for (i = 1, mask = 0; (0 == i)||(i < depth); i<<=1)
		mask += i;

	f(set_depth)(p, i);
	f(set_mask)(p, mask);
	
	f(set_width)(p, width);
	f(set_empty)(p, TRUE);
	f(set_full)(p, FALSE);
	f(set_fill)(p, 0);

	// The basic queue is an array of pointers
	f(set_queue)(p, malloc1D(f(get_depth)(p), sizeof(void *)));
	if (!f(get_queue)(p)) {f(delete)(p); return NULL;}
    f(set_read)(p, 0);
    f(set_write)(p, 0);

	// Initialize all pointers to NULL;
	for (i = 0; i < f(get_depth)(p); i++)
		f(get_queue)(p)[i] = NULL;

	// Each element in the array must now be made to point to a block
	// of memory of size 'width'. 

	// First, attempt to allocate sufficient memory to store all of the data
	// in a single block. If this fails, allocate separate blocks.

	f(set_block)(p, malloc1D(f(get_depth)(p), f(get_width)(p)));

	for (i = 0; i < f(get_depth)(p); i++)
	{
		if (f(get_block)(p)) // Was able to allocate as one block
			f(get_queue)(p)[i] = (char *)f(get_block)(p) + i*f(get_width)(p);
		else
			f(get_queue)(p)[i] = malloc1D(1, f(get_width)(p));
	}

	// Verify that all pointers were successfully initialized
	// NOTE: If a NULL pointer is found, the structure is deleted immediately.
	//       The loop test will then fail because get_depth() will return 0.
	for (i = 0; i < f(get_depth)(p); i++)
		if (!f(get_queue)(p)[i])
			p = f(delete)(p);

	return p;
}

STRUCT *f(flush)(STRUCT *p)
{
    f(set_read)(p, 0);
    f(set_write)(p, 0);

	f(set_empty)(p, TRUE);
	f(set_full)(p, FALSE);
	f(set_fill)(p, 0);

	return p;
}

int f(is_empty)(STRUCT *p)
{
	return f(get_empty)(p);
}

int f(is_full)(STRUCT *p)
{
	return f(get_full)(p);
}

int f(count)(STRUCT *p)
{
	return f(get_fill)(p);
}

int f(size)(STRUCT *p)
{
	return f(get_depth)(p);
}

int f(width)(STRUCT *p)
{
	return f(get_width)(p);
}

// If 'item' is NULL, then it is assumed that all that is being requested
// is the pointer to the newest item. The programmer is responsible 
// for transfering any data to it before any subsequent pops can retrieve
// the data. It is important that the value of the pointer not be retained.

// If 'item' is not NULL, then it is assumed that exactly 'width' bytes are
// to be copied from the item to the FIFO.

// NOTE: Pushing to a full FIFO will overwrite oldest data
void  *f(push)(STRUCT *p, void *item)
{
	void *newest;
	newest = f(get_queue)(p)[f(get_write)(p)];

	if (item)
		copy(item, newest, f(get_width)(p));

	if (!f(get_full)(p))
	{
		f(set_write)(p, (f(get_write)(p)+1)&f(get_mask)(p));
		f(set_fill)(p, f(get_fill)(p)+1);
	}

	f(set_empty)(p, FALSE);
	f(set_full)(p, f(get_read)(p)==f(get_write)(p));

	return newest;

}

// If 'item' is NULL, then it is assumed that all that is being requested
// is the pointer to the item to be read. The programmer is responsible 
// for fetching any data from it before any subsequent pushes can overwrite
// the data. It is important that the value of the pointer not be retained
// nor used to write data to the block.

// If 'item' is not NULL, then it is assumed that exactly 'width' bytes are
// to be copied from the FIFO to the item.

// NOTE: Popping from an empty FIFO will return last data again

void  *f(pop)(STRUCT *p, void *item)
{
	void *oldest;
	oldest = f(get_queue)(p)[f(get_read)(p)];

	if (item)
		copy(oldest, item, f(get_width)(p));

	if (!f(get_empty)(p))
	{
		f(set_read)(p, (f(get_read)(p)+1)&f(get_mask)(p));
		f(set_fill)(p, f(get_fill)(p)-1);
	}

	f(set_empty)(p, f(get_read)(p)==f(get_write)(p));
	f(set_full)(p, FALSE);

	return oldest;
}

int PFIFO_test(int test_number)
{
	PFIFO *fifo;
	char s[100], c;
	int i, len;

	fifo = PFIFO_new(NULL, 8, 1);
	if (!fifo)
		return EXIT_FAILURE;

	PFIFO_report(fifo);

	strcpy(s, "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
	len = strlen(s);

	i = 0;
	while (i < len)
	{
		if (!PFIFO_is_full(fifo))
			PFIFO_push(fifo, s+i++);
		else
		{
			while (!PFIFO_is_empty(fifo))
			{
				PFIFO_pop(fifo, &c);
				putc(c, stdout);
			}
			putc(' ', stdout);
		}
	}
	while (!PFIFO_is_empty(fifo))
	{
		PFIFO_pop(fifo, &c);
		putc(c, stdout);
	}
	putc(' ', stdout);

	PFIFO_report(fifo);

	PFIFO_delete(fifo);
	return EXIT_SUCCESS;
}

