/*
 * =======================================================================
 * FILE:        SDR.c
 * DESCRIPTION: Software Defined Radio structure
 * CREATED      24 MAR 10
 * MODIFIED     
 * PROGRAMMER:  William L. Bahn
 * COPYRIGHT:   2010
 *
 * These functions may be used for any purpose without royalty provided
 * the original author is acknowledged.
 * =======================================================================
 */

/*
 * =======================================================================
 * REVISION HISTORY
 *
 * REV 1.0: 24 MAR 10 Initial Creation.
 * =======================================================================
 */

//===========================================================================

//---------------------------------------------------------------------------
// Include Files
//---------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDR.h"

#define SDR_ERR_TYPES   (9)
#define SDR_ERR_UNKNOWN (0)
#define SDR_ERR_SDR     (1)
#define SDR_ERR_BLOCK   (2)
#define SDR_ERR_FIFO    (3)
#define SDR_ERR_PORT    (4)
#define SDR_ERR_MONITOR (5)
#define SDR_ERR_INI     (6)
#define SDR_ERR_RST     (7)
#define SDR_ERR_RUN     (8)
#define SDR_ERR_FIN     (9)

//---------------------------------------------------------------------------
// Structure Definition
//---------------------------------------------------------------------------

#define MAJOR_REV (1)
#define MINOR_REV (1)

#define STRUCT SDR
#define CODESTRING "SDR_"

struct STRUCT
{
	// Housekeeping Elements
	DWORD code;
	char *structname;
	// Live Elements
	int blocks;      // 0
	int fifos;       // 0
	int monitors;    // 0
	int errors;      // 0
	SDRB **block;    // NULL
	PFIFO **fifo;    // NULL
	SDRB **monitor;  // NULL
	void *data;      // NULL
	f(ModFunc) call; // NULL
	int err[SDR_ERR_TYPES+1];
};

#define CODE ((((((\
	(CODESTRING[0])<<8)\
	+CODESTRING[1])<<8)\
	+CODESTRING[2])<<8)\
	+CODESTRING[3])

#define SDR_ITEM_BLOCK   (0)
#define SDR_ITEM_FIFO    (1)
#define SDR_ITEM_MONITOR (2)

//===========================================================================
// Primitive Functions
//===========================================================================
//---------------------------------------------------------------------------
// Housekeeping
//---------------------------------------------------------------------------

DWORD f(get_code)(STRUCT *p)
{
	return (p)? p->code : 0;
}

DWORD f(set_code)(STRUCT *p)
{
	if (p)
		p->code = CODE;

	return f(get_code)(p);
}

int f(structcheck)(STRUCT *p)
{
	if ((p)&&(f(get_code)(p)==CODE))
		return 1;
	return 0;
}

char *f(get_structname)(STRUCT *p)
{
	return (f(structcheck)(p))? p->structname : NULL;
}

char *f(set_structname)(STRUCT *p)
{
	if (f(structcheck)(p))
		p->structname = VNQUOTES(STRUCT);
	return f(get_structname)(p);
}

//---------------------------------------------------------------------------
// Element Creation/Deletion
//---------------------------------------------------------------------------

STRUCT *f(create)(void)
{
	STRUCT *p;

	p = (STRUCT *) malloc(sizeof(STRUCT));
	f(set_code)(p);
	f(set_structname)(p);

	return p;
}

STRUCT *f(destroy)(STRUCT *p)
{
	if (f(structcheck)(p))
	{
		free(p);
		p = NULL;
	}

	return p;
}

//---------------------------------------------------------------------------
// Element Access
//---------------------------------------------------------------------------

getset(int,        blocks,   0);
getset(int,        fifos,    0);
getset(int,        monitors, 0);
getset(int,        errors,   0);
getset(SDRB**,     block,    NULL);
getset(PFIFO**,    fifo,     NULL);
getset(SDRB**,     monitor,  NULL);
getset(void*,      data,     NULL);
getset(f(ModFunc), call, NULL);

//===========================================================================
// Private Functions
//===========================================================================

// The private functions perform little to no parameter checking.
// The public functions calling them are responsible for this.

void f(report_elements)(STRUCT* p, FILE *fp)
{
	int i;

	char *errtype[] = 
		{"UNKNOWN", "SDR", "BLOCK", "FIFO", "PORT", "MONITORS", 
		 "INT", "RST", "RUN", "DEL"};

	fprintf(fp, "   blocks    = %i\n", f(get_blocks)(p));
	fprintf(fp, "   fifos     = %i\n", f(get_fifos)(p));
	fprintf(fp, "   monitors  = %i\n", f(get_monitors)(p));
	fprintf(fp, "   block     = %p\n", f(get_block)(p));
	fprintf(fp, "   fifo      = %p\n", f(get_fifo)(p));
	fprintf(fp, "   monitor   = %p\n", f(get_monitor)(p));
	fprintf(fp, "   data      = %p\n", f(get_data)(p));
	fprintf(fp, "   call      = %p\n", f(get_call)(p));
	fprintf(fp, "   errors    = %i\n", f(get_errors)(p));
	for (i = 0; i <= SDR_ERR_TYPES; i++)
		fprintf(fp, "   error[%02i] = %2i (%s)\n", i, p->err[i], errtype[i]);

}

void *f(add_item)(STRUCT *p, void *item, int type)
{
	switch (type)
	{
		case SDR_ITEM_BLOCK:
			f(get_block)(p)[f(get_blocks)(p)] = item;
			f(set_blocks)(p, f(get_blocks)(p)+1);
			break;
		case SDR_ITEM_FIFO:
			f(get_fifo)(p)[f(get_fifos)(p)] = item;
			f(set_fifos)(p, f(get_fifos)(p)+1);
			break;
		case SDR_ITEM_MONITOR:
			f(get_monitor)(p)[f(get_monitors)(p)] = item;
			f(set_monitors)(p, f(get_monitors)(p)+1);
			break;
	}
	
	return item;
}

void *f(get_item_n)(STRUCT *p, int n, int type)
{
	void *item;

	switch (type)
	{
		case SDR_ITEM_BLOCK:
			item = f(get_block)(p)[n];
			break;
		case SDR_ITEM_FIFO:
			item = f(get_fifo)(p)[n];
			break;
		case SDR_ITEM_MONITOR:
			item = f(get_monitor)(p)[n];
			break;
	}
	
	return item;
}

SDRB *f(get_block_n)(STRUCT *p, int n)
{
	return f(get_block)(p)[n];
}

SDRB *f(find_block)(STRUCT *p, char *name)
{
	int i;

	if (!name)
		return NULL;

	for (i = 0; i < f(get_blocks)(p); i++)
		if (SDRB_is_name(f(get_block)(p)[i], name))
			return f(get_block)(p)[i];

	return NULL;
}

int f(is_running)(STRUCT *p)
{
	int i, monitors;

	monitors = f(get_monitors)(p);
	for (i = 0; i < monitors; i++)
		if (SDRB_is_active(f(get_monitor)(p)[i]))
			return TRUE;

	return FALSE;
}

char *dstrcpy(char *s)
{
	char *d;
	int len;

	// Creates a dynamically allocated copy of a string.
	if (!s)
		return NULL;

	len = strlen(s);

	d = malloc(len+1);

	if (d)
		strcpy(d, s);

	return d;
}

int f(extract_port)(char *s)
{
	// Legal name forms: "name" or "name:port"
	// This function modifies the contents of s

	char *p;
	int port;

	// Set p to point to ':' delimiter
	for (p = s; ((*p)&&(':'!=*p)); p++)
		;

	port = 0; // default port
	if (*p) // delimiter found
	{
		*p = '\0';
		p++;
		port = atoi(p);
	}

	return port;
}

//===========================================================================
// Public Functions
//===========================================================================

WORD f(version)(int type)
{
	switch (type)
	{
		case STRUCT_VERSION: return (MAJOR_REV<<8)+MINOR_REV; break;
		case STRUCT_MAJOR:   return MAJOR_REV; break;
		case STRUCT_MINOR:   return MINOR_REV; break;
	}
	return 0;
}

void f(report)(STRUCT* p, FILE *fp)
{
	if (NULL == fp)
		fp = stdout;

	fprintf(fp, "\n");
	fprintf(fp, "================================================\n");
	fprintf(fp, "Structure Name: %s (Code: %lX)\n", 
		f(get_structname)(p), f(get_code)(p));
	fprintf(fp, "Version: (%u.%u)\n", 
		f(version)(STRUCT_MAJOR), f(version)(STRUCT_MINOR));
	fprintf(fp, "------------------------------------------------\n");
	fprintf(fp, "   Size: %i bytes\n", sizeof(STRUCT));
	fprintf(fp, "   Addr: %p\n", p);
	fprintf(fp, "------------------------------------------------\n");
	f(report_elements)(p, fp);
	fprintf(fp, "================================================\n");
}

//---------------------------------------------------------------------------
// Constructor/Destructor
//---------------------------------------------------------------------------

int f(delete)(STRUCT *p)
{
	int i;

	if (!p) return 1;

	for (i = 0; i < f(get_blocks)(p); i++)
		SDRB_delete(f(get_block)(p)[i]);

	for (i = 0; i < f(get_fifos)(p); i++)
		PFIFO_delete(f(get_fifo)(p)[i]);

	free(f(get_block)(p));
	free(f(get_fifo)(p));
	free(f(get_monitor)(p));

	free(p);

	return 0;
}

STRUCT *f(new)(void *data, f(ModFunc) call)
{
	STRUCT *p;
	int i;
	
	p = NULL;

	p = f(create)();
	if (!p)
		return NULL;

	f(set_errors)   (p, f(get_errors)(NULL));
	f(set_blocks)   (p, f(get_blocks)(NULL));
	f(set_fifos)    (p, f(get_fifos)(NULL));
	f(set_monitors) (p, f(get_monitors)(NULL));
	f(set_block)    (p, f(get_block)(NULL));
	f(set_fifo)     (p, f(get_fifo)(NULL));
	f(set_monitor)  (p, f(get_monitor)(NULL));
	f(set_data)     (p, f(get_data)(NULL));

	f(set_block)(p, malloc1D(20, sizeof(SDRB *)));
	f(set_fifo)(p, malloc1D(20, sizeof(PFIFO *)));
	f(set_monitor)(p, malloc1D(20, sizeof(SDRB *)));

	if (!f(get_block)(p))
	{
		f(delete)(p);
		p = NULL;
	}

	if (!f(get_fifo)(p))
	{
		f(delete)(p);
		p = NULL;
	}

	if (!f(get_monitor)(p))
	{
		f(delete)(p);
		p = NULL;
	}

	f(set_data)(p, data);

	for (i = 0; i <= SDR_ERR_TYPES; i++)
		p->err[i] = 0;

	return p;
}


//---------------------------------------------------------------------------
// Primary Functions
//---------------------------------------------------------------------------

int f(errors)(STRUCT *p)
{
	return f(get_errors)(p);
}

int f(log_error)(STRUCT *p, int type)
{
	if ((type < 0)||(type > SDR_ERR_TYPES))
		type = SDR_ERR_UNKNOWN;

	p->err[type]++;

	return f(set_errors)(p, f(get_errors)(p) + 1);
}

SDRB *f(add_block)(STRUCT *p, char *name, SDRB_ModFunc call)
{
	SDRB *blk;

	blk = f(add_item)(p, SDRB_new(name, NULL, call), SDR_ITEM_BLOCK);
	SDRB_set_radio(blk, p);

	if (SDRB_get_call(blk))
		(*SDRB_get_call(blk))(blk, SDRB_INI);

	if (NULL == blk)
		f(log_error)(p, SDR_ERR_BLOCK);

	return blk;
}

PFIFO *f(add_fifo)(STRUCT *p, PFIFO *fifo)
{
	fifo = f(add_item)(p, fifo, SDR_ITEM_FIFO);

	if (NULL == fifo)
		f(log_error)(p, SDR_ERR_FIFO);

	return fifo;
}

SDRB *f(add_monitor)(STRUCT *p, char *blockname)
{
	SDRB *blk;

	blk = f(add_item)(p, f(find_block)(p, blockname), SDR_ITEM_MONITOR);

	if (NULL == blk)
		f(log_error)(p, SDR_ERR_MONITOR);

	return blk;
}

PFIFO *f(connect)(STRUCT *p, char *from, char *to, int width, int depth)
{
	SDRB *from_block, *to_block;
	char *from_name, *to_name;
	int from_port, to_port;
	PFIFO *fifo;
	PORT *port;

	from_name = dstrcpy(from);
	to_name = dstrcpy(to);

	from_port = f(extract_port)(from_name);
	to_port   = f(extract_port)(to_name);

	from_block = f(find_block)(p, from_name);
	to_block   = f(find_block)(p, to_name);

	fifo = PFIFO_new(NULL, depth, width);
	fifo = f(add_fifo)(p, fifo);
	if (NULL == fifo)
		f(log_error)(p, SDR_ERR_FIFO);

	port = SDRB_new_port(from_block, PORT_OUTPUT, from_port, width, 0, fifo, NULL);
	if (NULL == port)
		f(log_error)(p, SDR_ERR_PORT);

	port = SDRB_new_port(to_block,   PORT_INPUT,  to_port,   width, 0, fifo, from_block);
	if (NULL == port)
		f(log_error)(p, SDR_ERR_PORT);

	free(from_name);
	free(to_name);

	return fifo;
}

int f(reset)(STRUCT *p)
{
	int i;

	if (!p)
	{
		f(log_error)(p, SDR_ERR_RST);
		return 1;
	}

	// Reset all of the SDR Blocks
	for (i = 0; i < f(get_blocks)(p); i++)
		SDRB_call(f(get_block)(p)[i], SDRB_RST);

	// Flush all interconnecting FIFOs
	for (i = 0; i < f(get_fifos)(p); i++)
		PFIFO_flush(f(get_fifo)(p)[i]);

	return 0;
}

int f(run)(STRUCT *p)
{
	int i;
	int was_active, is_active;
	int blocks;
	SDRB *block;

	while (f(is_running)(p))
	{
		blocks = f(get_blocks)(p); // Could change dynamically
		for (i = 0; i < blocks; i++)
		{
			block = f(get_block)(p)[i];
			was_active = is_active = SDRB_is_active(block);
			if (!SDRB_is_source(block))
				is_active = SDRB_mark_status(block);
			if (was_active || is_active)
				SDRB_call(block, SDRB_RUN);
		}
	}
	return 0;
}

void *f(datablock)(STRUCT *p)
{
	return f(get_data)(p);
}

//---------------------------------------------------------------------------
// Test Functions
//---------------------------------------------------------------------------

/*
#include "Source.h"
#include "Sink.h"

int f(test)(void)
{
	SDR *sdr;

	// Instantiate system
	sdr = SDR_new();

	// Define Blocks
	SDR_addblock(sdr, "Source", &Source_ini, &Source_run, &Source_del, NULL);
	SDR_addblock(sdr, "Sink",   &Sink_ini,   &Sink_run,   &Sink_del,   NULL);

	// Connect Blocks
	SDR_connect(sdr, "Source", 0, "Sink", 0, 16);

	// Tell system which block to monitor for end-of-run
	SDR_monitor(sdr, "Sink");
	
	// Run the system if no errors logged during construction
	SDR_run(sdr);

	// Delete the entire system
	SDR_delete(sdr);

	return EXIT_SUCCESS;
}
*/
