/*
 * =======================================================================
 * FILE:        SDRB.h
 * DESCRIPTION: Software Defined Radio Block structure
 * CREATED      21 MAR 10
 * MODIFIED     
 * PROGRAMMER:  William L. Bahn
 * COPYRIGHT:   2010
 *
 * These functions may be used for any purpose without royalty provided
 * the original author is acknowledged.
 * =======================================================================
 */

/*
 * =======================================================================
 * REVISION HISTORY
 *
 * REV 1.1: 25 MAR 10 Multi-port support added.
 * REV 1.0: 21 MAR 10 Initial Creation.
 * =======================================================================
 */

//===========================================================================

//---------------------------------------------------------------------------
// Include Files
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>

#include "SDRB.h"
#include "PORT.h"

#include "dirtyd.h"
#include "bytes.h"

//---------------------------------------------------------------------------
// Structure Definition
//---------------------------------------------------------------------------

#define MAJOR_REV (1)
#define MINOR_REV (1)

#define STRUCT SDRB
#define CODESTRING "SDRB"

struct STRUCT
{
	// Housekeeping Elements
	DWORD code;
	char *structname;
	// Live Elements
	char *name;      // NULL
	SDRB *self;      // NULL
	SDR *radio;      // NULL
	int source;      // FALSE
	int active;      // FALSE
	void *data;      // NULL
	int maxports;    // SDRB_PORTS
	f(ModFunc) call; // NULL

	PORT **in_port;  // NULL
	PORT **out_port; // NULL
};

#define CODE ((((((\
	(CODESTRING[0])<<8)\
	+CODESTRING[1])<<8)\
	+CODESTRING[2])<<8)\
	+CODESTRING[3])

//===========================================================================
// Primitive Functions
//===========================================================================
//---------------------------------------------------------------------------
// Housekeeping
//---------------------------------------------------------------------------

DWORD f(get_code)(STRUCT *p)
{
	return (p)? p->code : 0;
}

DWORD f(set_code)(STRUCT *p)
{
	if (p)
		p->code = CODE;

	return f(get_code)(p);
}

int f(structcheck)(STRUCT *p)
{
	if ((p)&&(f(get_code)(p)==CODE))
		return 1;
	return 0;
}

char *f(get_structname)(STRUCT *p)
{
	return (f(structcheck)(p))? p->structname : NULL;
}

char *f(set_structname)(STRUCT *p)
{
	if (f(structcheck)(p))
		p->structname = VNQUOTES(STRUCT);
	return f(get_structname)(p);
}

STRUCT *f(create)(void)
{
	STRUCT *p;

	p = (STRUCT *) malloc(sizeof(STRUCT));
	f(set_code)(p);
	f(set_structname)(p);

	return p;
}

STRUCT *f(destroy)(STRUCT *p)
{
	if (f(structcheck)(p))
	{
		free(p);
		p = NULL;
	}

	return p;
}

//---------------------------------------------------------------------------
// Element Access
//---------------------------------------------------------------------------

getset(char*,      name,      NULL); 
getset(SDRB*,      self,      NULL);
getset(SDR*,       radio,     NULL);
getset(int,        source,    FALSE);
getset(int,        active,    FALSE);
getset(int,        maxports,  SDRB_PORTS);
getset(void*,      data,      NULL);
getset(f(ModFunc), call,      NULL);

getset(PORT**,     in_port,   NULL);
getset(PORT**,     out_port,  NULL);

//===========================================================================
// Private Functions
//===========================================================================

// The private functions perform little to no parameter checking.
// The public functions calling them are responsible for this.

void f(report_elements)(STRUCT* p, FILE *fp)
{
	fprintf(fp, "   name     = %s\n", f(get_name)(p));
	fprintf(fp, "   self     = %p\n", f(get_self)(p));
	fprintf(fp, "   radio    = %p\n", f(get_radio)(p));
	fprintf(fp, "   source   = %i\n", f(get_source)(p));
	fprintf(fp, "   active   = %i\n", f(get_active)(p));
	fprintf(fp, "   maxports = %i\n", f(get_maxports)(p));
	fprintf(fp, "   data     = %i\n", f(get_data)(p));
	fprintf(fp, "   in_port  = %p\n", f(get_in_port)(p));
	fprintf(fp, "   out_port = %p\n", f(get_out_port)(p));
	fprintf(fp, "   call     = %p\n", f(get_call)(p));
}

//===========================================================================
// Public Functions
//===========================================================================

WORD f(version)(int type)
{
	switch (type)
	{
		case STRUCT_VERSION: return (MAJOR_REV<<8)+MINOR_REV; break;
		case STRUCT_MAJOR:   return MAJOR_REV; break;
		case STRUCT_MINOR:   return MINOR_REV; break;
	}
	return 0;
}

void f(report)(STRUCT* p, FILE *fp)
{
	if (NULL == fp)
		fp = stdout;

	fprintf(fp, "\n");
	fprintf(fp, "================================================\n");
	fprintf(fp, "Structure Name: %s (Code: %lX)\n", 
		f(get_structname)(p), f(get_code)(p));
	fprintf(fp, "Version: (%u.%u)\n", 
		f(version)(STRUCT_MAJOR), f(version)(STRUCT_MINOR));
	fprintf(fp, "------------------------------------------------\n");
	fprintf(fp, "   Size: %i bytes\n", sizeof(STRUCT));
	fprintf(fp, "   Addr: %p\n", p);
	fprintf(fp, "------------------------------------------------\n");
	f(report_elements)(p, fp);
	fprintf(fp, "================================================\n");
}

//---------------------------------------------------------------------------

int f(delete)(STRUCT *p)
{
	int i;

	if (!p) return 1;

	// Attached FIFOs are NOT deleted by this function.
	// They must be destroyed independently

	// Invoke Block-specific destructor
	if (f(get_call)(p))
		(*(f(get_call)(p)))(p, SDRB_DEL);

	for (i = 0; i < SDRB_PORTS; i++)
	{
		free(f(get_in_port)(p)[i]);
		free(f(get_out_port)(p)[i]);
	}
	free(f(get_in_port)(p));
	free(f(get_out_port)(p));

	free(p);
	return 0;
};

STRUCT *f(new)(char *name, void *data, f(ModFunc) call)
{
	int i;
	STRUCT *p;
	
	p = NULL;

	// Sanity Checks
	if (!name)
		return NULL;

	p = f(create)();
	if (!p)
		return NULL;

	f(set_name)      (p, f(get_name)(NULL));
	f(set_self)      (p, f(get_self)(NULL));
	f(set_radio)     (p, f(get_radio)(NULL));
	f(set_source)    (p, f(get_source)(NULL));
	f(set_active)    (p, f(get_active)(NULL));
	f(set_maxports)  (p, f(get_maxports)(NULL));
	f(set_data)      (p, f(get_data)(NULL));
	f(set_call)      (p, f(get_call)(NULL));

	f(set_in_port)   (p, f(get_in_port)(NULL));
	f(set_out_port)  (p, f(get_out_port)(NULL));

	f(set_name)      (p, name);
	f(set_self)      (p, p);
	f(set_source)    (p, TRUE);
	f(set_active)    (p, TRUE);
	f(set_data)      (p, data);
	f(set_call)      (p, call);

	f(set_in_port)(p, malloc1D(SDRB_PORTS, sizeof(PORT *)));
	f(set_out_port)(p, malloc1D(SDRB_PORTS, sizeof(PORT *)));

	for (i = 0; i < SDRB_PORTS; i++)
	{
		f(get_in_port)(p)[i] = NULL;
		f(get_out_port)(p)[i] = NULL;
	}

	return p;
}

int f(is_source)(STRUCT *p) // Is the block a source block (no input ports)
{
	return f(get_source)(p);
}

int f(is_name)(STRUCT *p, char *name)
{
	return (!strcmp(f(get_name)(p), name));
}

PORT *f(new_port)(STRUCT *p, int dir, int num, int width, int qty, PFIFO *fifo, SDRB *from)
{
	PORT *port;

	if (NULL == p)
		return NULL;
	if ((num < 0)||(num >= f(get_maxports)(p)))
		return NULL;

	port = PORT_new();

	PORT_set_dir(port, dir);
	PORT_set_width(port, width);
	PORT_set_qty(port, qty);
	PORT_set_fifo(port, fifo);
	PORT_set_from(port, from);

	switch (dir)
	{
		case PORT_INPUT:
			f(get_in_port)(p)[num] = port;
			f(set_source)(p, FALSE);
			break;
		case PORT_OUTPUT:
			f(get_out_port)(p)[num] = port;
			break;
	}
	return port;
}

int f(is_full)(STRUCT *p, int port) // Is an output port full
{
	PFIFO *fifo;

	fifo = PORT_get_fifo(SDRB_get_out_port(p)[port]);

	return PFIFO_is_full(fifo);
}

int f(is_empty)(STRUCT *p, int port) // Is an input port empty
{
	PFIFO *fifo;

	fifo = PORT_get_fifo(SDRB_get_in_port(p)[port]);

	return PFIFO_is_empty(fifo);
}

void *f(push)(STRUCT *p, int port, void *item)
{
	return PFIFO_push(PORT_get_fifo(f(get_out_port)(p)[port]), item);
}

void *f(pop)(STRUCT *p, int port, void *item)
{
	return PFIFO_pop(PORT_get_fifo(f(get_in_port)(p)[port]), item);
}

int f(is_active)(STRUCT *p) // Is the block marked active
{
	return f(get_active)(p);
}

int f(is_port_active)(STRUCT *p, int port) // Is an input port still alive
{
	// If block feeding it is alive, then port is alive
	if (SDRB_is_active(PORT_get_from(SDRB_get_in_port(p)[port])))
		return TRUE;

	// If block feeding it is dead, then port is alive if still has data
	return !SDRB_is_empty(p, port);
}

int f(is_any_port_active)(STRUCT *p) // Is any input port still alive
{
	int i;
	PORT *port;

	for (i = 0; i < SDRB_PORTS; i++)
	{
		port = f(get_in_port)(p)[i];
		if (port)
			if (f(is_port_active)(p, i))
				return TRUE;
	}

	// If this point is reached then all ports are inactive
	return FALSE;
}

int f(mark_status)(STRUCT *p)
{
	// Marks the block as active if it is a source block
	// or if it has any active ports.
	int active;

	active = f(is_source)(p)||f(is_any_port_active)(p);
	f(set_active)(p, active);
	
	return active;

}

int f(call)(STRUCT *p, int when)
{
	int retvalue;

	retvalue = 0; // Set retvalue to 1 to indicate errors.
	switch (when)
	{
		case SDRB_INI: // Initialation Behavior
			if (f(get_call)(p))
				retvalue = (*(f(get_call)(p)))(p, when);
			break;

		case SDRB_RST: // Reset Behavior
			f(set_active)(p, TRUE);
			if (f(get_call)(p))
				retvalue = (*(f(get_call)(p)))(p, when);
			break;

		case SDRB_RUN: // Runtime Behavior
			if (f(get_call)(p))
				retvalue = (*(f(get_call)(p)))(p, when);
			break;

		case SDRB_DEL: // Deletion Behavior
			if (f(get_call)(p))
				retvalue = (*(f(get_call)(p)))(p, when);
			break;

		default: // Should never reach this point.
			retvalue = 1;
	}

	return retvalue;

	return 0;
}

SDR *f(radio)(STRUCT *p)
{
	return f(get_radio)(p);
}