/*
 * ==========================================================================
 * FILE:        module.h
 * DESCRIPTION: Header File to support Structure Modules
 * CREATED      24 MAR 10
 * MODIFIED     
 * PROGRAMMER:  William L. Bahn
 * COPYRIGHT:   2010
 *
 * These functions may be used for any purpose without royalty provided
 * the original author is acknowledged.
 * ==========================================================================
 */

/*
 * ==========================================================================
 * REVISION HISTORY
 *
 * REV 1.0: 24 MAR 10 Initial Creation.
 * ==========================================================================
 */

//===========================================================================
#ifndef MODULE_DOT_H
#define MODULE_DOT_H

// This directive prevents the contents of the file from being included
// more than once, as this can create problems or, at the very least, 
// annoying warnings about macro redefinitions. This has to be mated with
// a corresponding #endif at the end of the file.

//===========================================================================
//== INCLUDE FILES ==========================================================
//===========================================================================

#include "dirtyd.h"

//===========================================================================
//== MACRO DEFINITIONS ======================================================
//===========================================================================

#define STRUCT_VERSION (0)
#define STRUCT_MAJOR   (1)
#define STRUCT_MINOR   (2)

//-------------------------------------------------------------------------

#define fff(x,y) x##_##y
#define ff(x,y) fff(x,y)
#define f(function) ff(STRUCT,function)

/* This macro works with the STRUCT macro definition to generate function
   names in structure modules so that they are unique. This works as
   follows:

   #define STRUCT FRED  // Name of structure in a structure module

   then functions can be defined like the following:

   int f(get_max)(STRUCT *p)
   {
      return 0;
   }

   expands to:

   int FRED_get_max(FRED *p)
   {
      return 0;
   }
*/
//-------------------------------------------------------------------------
// Autogenerates the get() and set() functions in a structure module
#define getset(TYPE, VARIABLE, DEFAULT)                                   \
                                                                          \
TYPE f(get_##VARIABLE)(STRUCT *p)                                         \
{                                                                         \
	return (f(structcheck(p)))? p->VARIABLE : DEFAULT;                    \
}                                                                         \
                                                                          \
TYPE f(set_##VARIABLE)(STRUCT *p, TYPE v)                                 \
{                                                                         \
	if (f(structcheck)(p))                                                \
		p->VARIABLE = v;                                                  \
	return f(get_##VARIABLE)(p);                                          \
}                                                                         \
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
// Autogenerates the get() and set() functions prototypes in the header
// file for a structure. Requires the STRUCT be defined and, most likely
// undefined. Note that, normally, the get()/set() functions are not made
// publicly available, but in some instances it makes sense to do so.
#define getset_proto(TYPE, VARIABLE, DEFAULT)                             \
                                                                          \
TYPE f(get_##VARIABLE)(STRUCT *p);                                        \
TYPE f(set_##VARIABLE)(STRUCT *p, TYPE v);                                \
//-------------------------------------------------------------------------

#define NQUOTES(x) #x
#define VNQUOTES(x) NQUOTES(x)
#define debug(x, s) printf("%s:%u: %s=" s, __FILE__, __LINE__, #x, (x))

/* 
Macro expansion does not occur within, however the "stringizing" operator
can used to quote a macro after expansion has occured, thus:

printf("%s\n", NQUOTES(2+2));

expands to

printf("%s\n", "2+2");

and

debug(bob, "%i");

expands to

printf("%s:%u: %s=" "%i", __FILE__, __LINE__, #bob, (bob));

which, because back-to-back strings are concatenated, reduces to

printf("%s:%u: %s=%i", __FILE__, __LINE__, "bob", (bob));

The __FILE__ and __LINE__ are replaced by the file name and line numbers.

But if the desire is to put quotes around the value that a macro reduces to,
instead of the macro name, the # directive can be used indirectly as follows:

VNQUOTES(__LINE__)

reduces to

NQUOTES(42)

which reduces to

"42"

if it happens to be on line 42.
*/
//===========================================================================

#endif

